/*
 * GGPS: GPS satellite information display using GTK+ toolkit
 * Copyright 1999 Eric Smith <eric@brouhaha.com>
 *
 * $Id: ggps.c,v 1.6 1999/07/15 22:35:31 eric Exp $
 */

#include <stdio.h>
#include <math.h>
#include <gtk/gtk.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "ggps.h"
#include "nmea.h"


#define SKY_SIZE 360
#define SKY_BORDER 30
#define WINDOW_SIZE (SKY_SIZE + 2 * SKY_BORDER)


int fd = 0;


/* --- Backing pixmap for drawing area  --- */
static GdkPixmap *pixmap = NULL;
GdkFont *font = NULL;
GtkWidget *sat_clist;


#define DEG_TO_RAD(x) ((x)*M_PI/180)


#define SAT_CLIST_COLS 5
char *sat_clist_titles[SAT_CLIST_COLS] =
{ "PRN", "azimuth", "elevation", "SNR", "SNR" };

void set_sat_clist (void)
{
  int sat;
  char prn [3];
  char az [4];
  char el [3];
  char snr [3];
  char *row[SAT_CLIST_COLS] = { & prn [0], & az [0], & el [0], & snr [0], NULL };

  gtk_clist_clear (GTK_CLIST (sat_clist));
  for (sat = 0; sat < sats; sat++)
    {
      sprintf (prn, "%02d", sat_info [sat].prn);
      sprintf (az, "%d", sat_info [sat].azimuth);
      sprintf (el, "%d", sat_info [sat].elevation);
      sprintf (snr, "%d", sat_info [sat].snr);
      gtk_clist_append (GTK_CLIST (sat_clist), row);
    }
}


#define TEXT_BORDER 4

/*
 * Repaint
 *
 * data - widget to repaint
 */
gint Repaint (gpointer data)
{
    GtkWidget*    drawing_area = (GtkWidget *) data;
    GdkRectangle  update_rect;
    int midx, midy;
    int tw, th;
    int sat;
    int sky_radius = drawing_area->allocation.width/2 - SKY_BORDER;

    sat_info_updated = 0;

    if (! font)
      font = gdk_font_load ("fixed");  /* $$$ how should we pick a font? */

    /* --- clear pixmap (background image) --- */
    gdk_draw_rectangle (pixmap,
              drawing_area->style->white_gc,
              TRUE,
              0, 0,
              drawing_area->allocation.width,
              drawing_area->allocation.height);

    /* --- Calculate midpoint of clock --- */
    midx = drawing_area->allocation.width / 2;
    midy = drawing_area->allocation.height / 2;

    /* --- Draw Circle --- */
    gdk_draw_arc (pixmap, 
                  drawing_area->style->black_gc,
                  0,
                  SKY_BORDER, SKY_BORDER,
		  drawing_area->allocation.width - 2 * SKY_BORDER,
		  drawing_area->allocation.height - 2 * SKY_BORDER,
		  0, 360 * 64);

    for (sat = 0; sat < sats; sat++)
      {
        double rho = DEG_TO_RAD (90 - sat_info [sat].azimuth);
#if 1
        double theta = (90 - sat_info [sat].elevation) / 90.0;
#else
        double theta = cos (DEG_TO_RAD (sat_info [sat].elevation));
#endif
        int x = midx + cos (rho) * theta * sky_radius;
        int y = midy - sin (rho) * theta * sky_radius;
	char name [3];

	sprintf (name, "%02d", sat_info [sat].prn);
	tw = gdk_text_width  (font, name, strlen (name));
	th = gdk_text_height (font, name, strlen (name));
	gdk_draw_rectangle (pixmap,
			    drawing_area->style->black_gc,
			    sat_info [sat].snr < 0, /* fill */
			    x - tw/2 - TEXT_BORDER,
			    y - th/2 - TEXT_BORDER,
			    tw + 2 * TEXT_BORDER,
			    th + 2 * TEXT_BORDER);
	gdk_draw_string (pixmap, 
			 font,
			 (sat_info [sat].snr < 0) ?
			     drawing_area->style->white_gc :
			     drawing_area->style->black_gc,
			 x - tw/2,
			 y + th/2,
			 name);
      }

    /* --- Set the rect up - 
     *     The entire pixmap is going to be copied
     *     onto the window so the rect is configured 
     *     as the size of the window.
     */
 
    update_rect.x = 0;
    update_rect.y = 0;
    update_rect.width = drawing_area->allocation.width;
    update_rect.height = drawing_area->allocation.height;

    /* --- And then draw it (calls expose event) --- */
    gtk_widget_draw (drawing_area, &update_rect);

    return (TRUE);
}

/* 
 * configure_event
 *
 * Create a new backing pixmap of the appropriate size 
 * Of course, this is called whenever the window is 
 * resized.  We have to free up things we allocated.
 */
static gint configure_event (GtkWidget *widget, GdkEventConfigure *event)
{
    /* --- Free background if we created it --- */
    if (pixmap) {
        gdk_pixmap_unref (pixmap);
    } 

    /* --- Create a new pixmap with new size --- */
    pixmap = gdk_pixmap_new (widget->window,
                widget->allocation.width,
                widget->allocation.height,
                -1);

    return TRUE;
}

/*
 * expose_event
 *
 * When the window is exposed to the viewer or 
 * the gdk_widget_draw routine is called, this 
 * routine is called.  Copies the background pixmap
 * to the window.
 */
gint expose_event (GtkWidget *widget, GdkEventExpose *event)
{

    /* --- Copy pixmap to the window --- */
    gdk_draw_pixmap (widget->window,
          widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
          pixmap,
          event->area.x, event->area.y,
          event->area.x, event->area.y,
          event->area.width, event->area.height);

    return FALSE;
}


/*
 * quit
 *
 * Get out of the application
 */
void quit ()
{
    gtk_exit (0);
}


void repaint_sats (void *d)
{
  GtkWidget *sat_drawing_area = (GtkWidget *) d;
  set_sat_clist ();

  if (! sat_drawing_area)
    {
      fprintf (stderr, "where'd the sat_drawing_area go?\n");
      return;
    }
  Repaint (sat_drawing_area);
}


GtkWidget *create_page (GtkWidget *notebook, char *name)
{
  GtkWidget *label, *page;

  label = gtk_label_new (name);
  gtk_widget_show (label);
  page = gtk_frame_new (NULL);
  gtk_widget_show (page);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), page, label);
  return (page);
}


/*
 * main
 *
 * Program begins here.
 */ 
int main (int argc, char *argv[])
{
  int i;
  GtkWidget *window, *hbox, *sat_drawing_area;
  GtkWidget *notebook;
  GtkWidget *sat_page, *sat_box;

  gtk_init (&argc, &argv);

  if (argc > 2)
    {
      fprintf (stderr, "too many args\n");
      return (1);
    }

  fd = 0;  /* assume standard input */
  if ((argc == 2) && (strcmp (argv [1], "-") != 0))
    {
      fd = open (argv [1], O_RDONLY);
      if (fd < 0)
	{
	  fprintf (stderr, "open failed\n");
	  return (2);
	}
    }

  i = fcntl (fd, F_SETFL, O_NONBLOCK);
  if (i < 0)
    {
      fprintf (stderr, "can't set input to non-blocking\n");
      return (3);
    }

    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), hbox);
    gtk_widget_show (hbox);

    gtk_signal_connect (GTK_OBJECT (window), "destroy",
                GTK_SIGNAL_FUNC (quit), NULL);

    /* create notebook */
    notebook = gtk_notebook_new ();
#if 0
    gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
			GTK_SIGNAL_FUNC (PageSwitch), NULL);
#endif
    gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
    gtk_box_pack_start (GTK_BOX (hbox), notebook, TRUE, TRUE, 0);
    gtk_container_border_width (GTK_CONTAINER (notebook), 10);

    /* create satellite page */
    sat_page = create_page (notebook, "Satellites");

    /* create box for sat page */
    sat_box = gtk_hbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (sat_page), sat_box);
    gtk_widget_show (sat_box);

    /* --- Create the drawing area  --- */
    sat_drawing_area = gtk_drawing_area_new ();
    gtk_drawing_area_size (GTK_DRAWING_AREA (sat_drawing_area),
			   WINDOW_SIZE, WINDOW_SIZE);
    gtk_box_pack_start (GTK_BOX (sat_box), sat_drawing_area, TRUE, TRUE, 0);

    gtk_widget_show (sat_drawing_area);

    /* create the clist */
    sat_clist = gtk_clist_new_with_titles (5, sat_clist_titles);
    gtk_clist_set_shadow_type (GTK_CLIST(sat_clist), GTK_SHADOW_OUT);

    gtk_box_pack_start (GTK_BOX (sat_box), sat_clist, TRUE, TRUE, 0);
    gtk_widget_show (sat_clist);

    /* --- Signals used to handle backing pixmap --- */
    gtk_signal_connect (GTK_OBJECT (sat_drawing_area), "expose_event",
                (GtkSignalFunc) expose_event, NULL);
    gtk_signal_connect (GTK_OBJECT(sat_drawing_area),"configure_event",
                (GtkSignalFunc) configure_event, NULL);

    /* --- Show the window --- */
    gtk_widget_show_all (window);

    repaint_sats (sat_drawing_area);

    nmea_init (fd, & repaint_sats, sat_drawing_area);

    /* --- Call gtk-main loop --- */
    gtk_main ();

    return 0;
}
