/*
 * GGPS: GPS satellite information display using GTK+ toolkit
 * Copyright 1999 Eric Smith <eric@brouhaha.com>
 *
 * I/O routines and NMEA sentence parsing
 *
 * $Id: nmea.c,v 1.5 1999/07/10 05:45:55 eric Exp $
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdk.h>

#include "nmea.h"


rs_callback repaint_sat_callback;
void *repaint_sat_callback_data;


int nmea_debug = 0;


static int nmea_checksum (char *s, int l)
{
  int c = 0;
  while (l--)
    c ^= *(s++);
  return (c);
}


/*
 * Extract an integer value from an NMEA sentence.
 * On entry, the string should start with the comma preceding the field.
 * Returns 0 for a null field,
 *         1 if a valid integer was found,
 *         -1 for an error or end of string
 */
static int get_int (char **s, int *r)
{
  char *p;

  if (**s == '\0')
    return (-1);
  if (**s != ',')
    return (-1);
  (*s)++;
  if ((**s == ',') || (**s == '\0'))
    {
      /* empty field */
      *r = 0;
      return (0);
    }
  *r = strtol (*s, &p, 10);
  if (p == *s)
    return (-1);
  *s = p;
  if ((**s != ',') && (**s != '\0'))
    return (-1);
  return (1);
}


int sat_info_updated = 0;
int sats = 0;
t_sat_info sat_info [MAX_SAT];

static int new_sats = 0;
static t_sat_info new_sat_info [MAX_SAT];


static int gpgsv_total_expected = 0;
static int gpgsv_next_expected = 0;
static int gpgsv_sats_expected = 0;

static void gpgsv_resync (void)
{
  gpgsv_total_expected = 0;
  gpgsv_next_expected = 0;
  gpgsv_sats_expected = 0;
  new_sats = 0;
  memset (& new_sat_info, 0, sizeof (new_sat_info));
}

static int parse_nmea_gpgsv (char *s)
{
  int tot, seq, sat_exp;

  if (get_int (& s, & tot) <= 0)
    return (-1);
  if (get_int (& s, & seq) <= 0)
    return (-1);
  if (get_int (& s, & sat_exp) <= 0)
    return (-1);
  if (gpgsv_total_expected == 0)
    {
      gpgsv_total_expected = tot;
      gpgsv_next_expected = 1;
      gpgsv_sats_expected = sat_exp;
    }
  else if ((gpgsv_total_expected != tot) ||
	   (gpgsv_next_expected != seq) ||
	   (gpgsv_sats_expected != sat_exp))
    {
      gpgsv_resync ();
      return (0);
    }

  while (*s == ',')
    {
      int prn_pr, el_pr, az_pr, snr_pr;
      prn_pr = get_int (& s, & new_sat_info [new_sats].prn);
      if (prn_pr < 0)
	return (-1);
      el_pr = get_int (& s, & new_sat_info [new_sats].elevation);
      if (el_pr < 0)
	return (-1);
      az_pr =  get_int (& s, & new_sat_info [new_sats].azimuth);
      if (az_pr < 0)
	return (-1);
      snr_pr = get_int (& s, & new_sat_info [new_sats].snr);
      if (snr_pr < 0)
	return (-1);
      if (prn_pr == 0)
	continue;
      if ((el_pr == 0) || (az_pr == 0))
	return (-1);  /* if the satellite is listed, it must have an
			 elevation and azimuth */
      if (snr_pr == 0)
	new_sat_info [new_sats].snr = -1;
#if 0
      printf ("sat %d: prn %d el %d az %d snr %d\n",
	      new_sats,
	      new_sat_info [new_sats].prn,
	      new_sat_info [new_sats].elevation,
	      new_sat_info [new_sats].azimuth,
	      new_sat_info [new_sats].snr);
#endif
      new_sats++;
    }

  gpgsv_next_expected++;

  if (gpgsv_next_expected > gpgsv_total_expected)
    {
      if (new_sats != gpgsv_sats_expected)
	printf ("expected %d satellites, got %d\n",
		gpgsv_sats_expected, new_sats);
      memcpy (sat_info, new_sat_info, sizeof (sat_info));
      sats = new_sats;
      gpgsv_resync ();
      sat_info_updated = 1;
      if (repaint_sat_callback)
	repaint_sat_callback (repaint_sat_callback_data);
    }

  return (0);
}

static int parse_nmea_gpgga (char *s)
{
  return (0);
}

static int parse_nmea_gpgsa (char *s)
{
  return (0);
}

static int parse_nmea_gprmc (char *s)
{
  return (0);
}

static int parse_nmea_pgrme (char *s)
{
  /* garmin proprietary */
  return (0);
}

static int parse_nmea_pgrmf (char *s)
{
  /* garmin proprietary */
  return (0);
}

static int parse_nmea_pgrmt (char *s)
{
  /* garmin proprietary */
  return (0);
}


typedef struct
{
  char *s;
  int (*fn)(char *s);
} t_nmea_sentence_info;

static t_nmea_sentence_info nmea_sentence_info [] =
{
  { "GPGSV", parse_nmea_gpgsv },
  { "GPGGA", parse_nmea_gpgga },
  { "GPGSA", parse_nmea_gpgsa },
  { "GPRMC", parse_nmea_gprmc },
  { "PGRME", parse_nmea_pgrme },
  { "PGRMF", parse_nmea_pgrmf },
  { "PGRMT", parse_nmea_pgrmt },
  { NULL, NULL }
};


static int parse_nmea_sentence (char *s)
{
  char *orig;
  int l;
  int c1, c2;
  t_nmea_sentence_info *si;

  l = strlen (s);

  /* throw away trailing CRs and LFs */
  while ((s [l-1] == '\r') || (s [l-1] == '\n'))
    {
      s [l-1] = '\0';
      l--;
    }

  orig = s;

  if (*s != '$')
    {
      if (nmea_debug)
	printf ("dollar sign missing, ");
      goto bad;
    }
  s++;  /* throw away the dollar sign */
  l--;

  if (s [l - 3] != '*')
    {
      if (nmea_debug)
	printf ("asterisk missing, ");
      goto bad;
    }
  c1 = nmea_checksum (& s [0], l - 3);
  c2 = strtol (& s [l - 2], NULL, 16);
  if (c1 != c2)
    {
      if (nmea_debug)
	printf ("bad checksum, computed %02x: ", c1);
      goto bad;
    }

  s [l - 3] = '\0';  /* now that it's been verified, throw away the checksum */
  l -= 3;

  for (si = & nmea_sentence_info [0]; si->s; si++)
    if (strncmp (& s [0], si->s, strlen (si->s)) == 0)
      {
	if (nmea_debug >= 2)
	  printf ("NMEA sentence %s\n", orig);
	return (si->fn (s + strlen (si->s)));
      }

  if (nmea_debug)
    printf ("unrecognized ");
  goto bad;

 bad:
  if (nmea_debug)
    printf ("NMEA sentence %s\n", orig);
  return (-1);
}


#define MAX_NMEA_SENTENCE_LENGTH 200


#define SER_BUF_SIZE 10000
static char ser_buf [SER_BUF_SIZE+1];  /* leave room for a NULL */
static int ser_buf_cnt = 0;  /* chars in buffer */


static void discard_chars (int c)
{
  memmove (& ser_buf [0], & ser_buf [c], ser_buf_cnt - c);
  ser_buf_cnt -= c;
}

static void nmea_read_data (gpointer data, gint source, GdkInputCondition condition)
{
  int cnt;
  char *p;

  for (;;)
    {
      cnt = read (source, & ser_buf [ser_buf_cnt], SER_BUF_SIZE - ser_buf_cnt);
      if (cnt < 0)
	{
	  /* $$$ error, what do we do? */
	  return;
	}
      if (cnt == 0)
	{
	  /* no data yet, try again later */
	  return;
	}
      if (cnt > SER_BUF_SIZE - ser_buf_cnt)
	{
	  printf ("how'd that happen?  cnt=%d\n", cnt);
	  cnt = SER_BUF_SIZE - ser_buf_cnt;
	}
      ser_buf_cnt += cnt;
      ser_buf [ser_buf_cnt] = '\0';  /* make sure our calls to strchr don't
					run off the end */
      if (ser_buf [0] != '$')
	{
	  p = strchr (& ser_buf [0], '$');
	  if (! p)
	    {
	      /* no dollar sign found, everything is junk */
	      ser_buf_cnt = 0;
	      return;
	    }
	  discard_chars (p - & ser_buf [0]);
	}
      p = strchr (& ser_buf [0], '\n');
      if (! p)
	{
	  /* haven't got a complete sentence yet */
	  if (ser_buf_cnt > MAX_NMEA_SENTENCE_LENGTH)
	    {
	      /* too much junk, throw some away */
	      discard_chars (MAX_NMEA_SENTENCE_LENGTH);
	    }
	  return;
	}
      *p = '\0';
      parse_nmea_sentence (& ser_buf [0]);
      discard_chars (p + 1 - & ser_buf [0]);
    }
}


/* fd is the Unix file descriptor to read NMEA data from */
void nmea_init (int fd,
		rs_callback rsc,
		void *rscd)
{
  gdk_input_add (fd, GDK_INPUT_READ, nmea_read_data, NULL);
  repaint_sat_callback = rsc;
  repaint_sat_callback_data = rscd;
}
